//
//  Pianobar.m
//  pianobarCocoa
//
//  Created by Remi Bernotavicius on 8/25/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "Pianobar.h"

@implementation Pianobar

/*	player is done, clean up
 */
static void BarMainPlayerCleanup (BarApp_t *app, pthread_t *playerThread) {
	void *threadRet;
    
	BarUiStartEventCmd (&app->settings, "songfinish", app->curStation,
                        app->playlist, &app->player, app->ph.stations, PIANO_RET_OK,
                        WAITRESS_RET_OK);
    
	/* FIXME: pthread_join blocks everything if network connection
	 * is hung up e.g. */
	pthread_join (*playerThread, &threadRet);
    
	/* don't continue playback if thread reports error */
	if (threadRet != (void *) PLAYER_RET_OK) {
		app->curStation = NULL;
	}
    
	memset (&app->player, 0, sizeof (app->player));
}

/*	start new player thread
 */
static void BarMainStartPlayback (BarApp_t *app, pthread_t *playerThread) {
	BarUiPrintSong (&app->settings, app->playlist, app->curStation->isQuickMix ?
                    PianoFindStationById (app->ph.stations,
                                          app->playlist->stationId) : NULL);
    
	if (app->playlist->audioUrl == NULL) {
		BarUiMsg (&app->settings, MSG_ERR, "Invalid song url.\n");
	} else {
		/* setup player */
		memset (&app->player, 0, sizeof (app->player));
        
		WaitressInit (&app->player.waith);
		WaitressSetUrl (&app->player.waith, app->playlist->audioUrl);
        
		/* set up global proxy, player is NULLed on songfinish */
		if (app->settings.proxy != NULL) {
			WaitressSetProxy (&app->player.waith, app->settings.proxy);
		}
        
		app->player.gain = app->playlist->fileGain;
		app->player.scale = BarPlayerCalcScale (app->player.gain + app->settings.volume);
		app->player.audioFormat = app->playlist->audioFormat;
		app->player.settings = &app->settings;
        pthread_mutex_init (&app->player.pauseMutex, NULL);
        
		/* throw event */
		BarUiStartEventCmd (&app->settings, "songstart",
                            app->curStation, app->playlist, &app->player, app->ph.stations,
                            PIANO_RET_OK, WAITRESS_RET_OK);
        
		/* prevent race condition, mode must _not_ be FREED if
		 * thread has been started */
		app->player.mode = PLAYER_STARTING;
		/* start player */
		pthread_create (playerThread, NULL, BarPlayerThread,
                        &app->player);
	}
}

/*	fetch new playlist
 */
static void BarMainGetPlaylist (BarApp_t *app) {
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataGetPlaylist_t reqData;
	reqData.station = app->curStation;
	reqData.quality = app->settings.audioQuality;
    
	BarUiMsg (&app->settings, MSG_INFO, "Receiving new playlist... ");
	if (!BarUiPianoCall (app, PIANO_REQUEST_GET_PLAYLIST,
                         &reqData, &pRet, &wRet)) {
		app->curStation = NULL;
	} else {
		app->playlist = reqData.retPlaylist;
		if (app->playlist == NULL) {
			BarUiMsg (&app->settings, MSG_INFO, "No tracks left.\n");
			app->curStation = NULL;
		}
	}
	BarUiStartEventCmd (&app->settings, "stationfetchplaylist",
                        app->curStation, app->playlist, &app->player, app->ph.stations,
                        pRet, wRet);
}


-(id)initWithUsername:(NSString*)username andPassword:(NSString*)password quality:(PianoAudioFormat_t)quality error:(NSString**)_error
{
    
//	if((self = [super init]))
//	{
//		NSString *resourcePath = [[[[NSBundle mainBundle] bundlePath] stringByAppendingPathComponent:@"Contents"] stringByAppendingPathComponent:@"MacOS"];
//		[[NSFileManager defaultManager] changeCurrentDirectoryPath:resourcePath];
//		
//		
//		currentStation = -1;
//		playing = NO;
//		ctlFd = NULL;
//		buf = '\0';
//		
//		memset (&app, 0, sizeof (app));
//
//		/* init some things */
//		ao_initialize ();
//		gnutls_global_init ();
//		PianoInit (&app.ph);
//		
//		WaitressInit (&app.waith);
//		app.waith.url.host = strdup (PIANO_RPC_HOST);
//		//app.waith.url.tls = true;
//		app.waith.tlsFingerprint = app.settings.tlsFingerprint;
//		
//		BarSettingsInit (&app.settings);
//		BarSettingsRead (&app.settings);
//        
//        //set the quality
//        if(quality == 0)
//        {
//            /* apply defaults */
//            #ifdef ENABLE_FAAD
//                app.settings.audioFormat = PIANO_AF_AACPLUS;
//            #else
//                #ifdef ENABLE_MAD
//                    app.settings.audioFormat = PIANO_AF_MP3;
//                #endif
//            #endif
//        }
//        else
//            app.settings.audioFormat = quality;
//		
//		/* init fds */
//		FD_ZERO(&app.input.set);
//		app.input.fds[0] = STDIN_FILENO;
//		FD_SET(app.input.fds[0], &app.input.set);
//		
//		/* open fifo read/write so it won't EOF if nobody writes to it */
//		assert (sizeof (app.input.fds) / sizeof (*app.input.fds) >= 2);
//		app.input.fds[1] = open (app.settings.fifo, O_RDWR);
//		if (app.input.fds[1] != -1) {
//			FD_SET(app.input.fds[1], &app.input.set);
//			BarUiMsg (&app.settings, MSG_INFO, "Control fifo at %s opened\n",
//					  app.settings.fifo);
//		}
//		app.input.maxfd = app.input.fds[0] > app.input.fds[1] ? app.input.fds[0] :
//		app.input.fds[1];
//		++app.input.maxfd;
//		
//		app.settings.username = strdup ([username cStringUsingEncoding:NSASCIIStringEncoding]);
//		app.settings.password = strdup ([password cStringUsingEncoding:NSASCIIStringEncoding]);
//		
//		/* set up proxy (control proxy for non-us citizen or global proxy for poor
//		 * firewalled fellows) */
//		if (app.settings.controlProxy != NULL) {
//			/* control proxy overrides global proxy */
//			WaitressSetProxy (&app.waith, app.settings.controlProxy);
//		} else if (app.settings.proxy != NULL && strlen (app.settings.proxy) > 0) {
//			WaitressSetProxy (&app.waith, app.settings.proxy);
//		}
//		
//		PianoReturn_t pRet;
//		WaitressReturn_t wRet;
//		PianoRequestDataLogin_t reqData;
//		reqData.user = app.settings.username;
//		reqData.password = app.settings.password;
//		reqData.step = 0;
//		
//		//NSLog(@"Login...");
//		
//		const char* error;
//		
//		if (!BarUiPianoCallError (&app, PIANO_REQUEST_LOGIN, &reqData, &pRet, &wRet, &error))
//		{
//			*_error = [NSString stringWithCString:error encoding:NSASCIIStringEncoding];
//			return self;
//		}
//		
//		PianoReturn_t pRet2;
//		WaitressReturn_t wRet2;
//		
//		if (!BarUiPianoCallError (&app, PIANO_REQUEST_GET_STATIONS, NULL, &pRet2, &wRet2, &error))
//		{
//			*_error = [NSString stringWithCString:error encoding:NSASCIIStringEncoding];
//			return self;
//		}
//		
//	}
//	
//	return self;
    
    
    
    if(self = [super init])
    {
        
        //Set some things for the class
        NSString *resourcePath = [[[[NSBundle mainBundle] bundlePath] stringByAppendingPathComponent:@"Contents"] stringByAppendingPathComponent:@"MacOS"];
        [[NSFileManager defaultManager] changeCurrentDirectoryPath:resourcePath];
        
        currentStation = -1;
        playing = NO;
        
        //setup pianobar
        
        memset (&app, 0, sizeof (app));
        
        /* init some things */
        ao_initialize ();
        gcry_check_version (NULL);
        gcry_control (GCRYCTL_DISABLE_SECMEM, 0);
        gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
        gnutls_global_init ();
        
        BarSettingsInit (&app.settings);
        BarSettingsRead (&app.settings);
        
        PianoInit (&app.ph, app.settings.partnerUser, app.settings.partnerPassword,
                   app.settings.device, app.settings.inkey, app.settings.outkey);
        
        
        //set some things for pianobar from the class' settings
        
        //set the quality
        if(quality == 0)
        {
            #ifdef ENABLE_FAAD
                app.settings.audioQuality = PIANO_AF_AACPLUS;
            #else
                #ifdef ENABLE_MAD
                    app.settings.audioQuality = PIANO_AF_MP3;
                #endif
            #endif
        }
        else
            app.settings.audioQuality = quality;
        
        app.settings.username = strdup ([username cStringUsingEncoding:NSASCIIStringEncoding]);
        app.settings.password = strdup ([password cStringUsingEncoding:NSASCIIStringEncoding]);
        
        //continue setting up pianobar
        
        BarUiMsg (&app.settings, MSG_NONE,
                  "Welcome to " PACKAGE " (" VERSION ")! ");
        if (app.settings.keys[BAR_KS_HELP] == BAR_KS_DISABLED) {
            BarUiMsg (&app.settings, MSG_NONE, "\n");
        } else {
            BarUiMsg (&app.settings, MSG_NONE,
                      "Press %c for a list of commands.\n",
                      app.settings.keys[BAR_KS_HELP]);
        }
        
        WaitressInit (&app.waith);
        app.waith.url.host = strdup (PIANO_RPC_HOST);
        app.waith.tlsFingerprint = app.settings.tlsFingerprint;
        
        /* init fds */
        FD_ZERO(&app.input.set);
        app.input.fds[0] = STDIN_FILENO;
        FD_SET(app.input.fds[0], &app.input.set);
        
        /* open fifo read/write so it won't EOF if nobody writes to it */
        assert (sizeof (app.input.fds) / sizeof (*app.input.fds) >= 2);
        app.input.fds[1] = open (app.settings.fifo, O_RDWR);
        if (app.input.fds[1] != -1) {
            FD_SET(app.input.fds[1], &app.input.set);
            BarUiMsg (&app.settings, MSG_INFO, "Control fifo at %s opened\n",
                      app.settings.fifo);
        }
        app.input.maxfd = app.input.fds[0] > app.input.fds[1] ? app.input.fds[0] : app.input.fds[1];
        ++app.input.maxfd;
        
        //set up the proxy
        
        /* set up proxy (control proxy for non-us citizen or global proxy for poor
         * firewalled fellows) */
        if (app.settings.controlProxy != NULL) {
            /* control proxy overrides global proxy */
            WaitressSetProxy (&app.waith, app.settings.controlProxy);
        } else if (app.settings.proxy != NULL && strlen (app.settings.proxy) > 0) {
            WaitressSetProxy (&app.waith, app.settings.proxy);
        }
        
        //log the user in
        
        PianoReturn_t pRet;
        WaitressReturn_t wRet;
        PianoRequestDataLogin_t reqData;
        reqData.user = app.settings.username;
        reqData.password = app.settings.password;
        reqData.step = 0;
        
        const char* error;
        
        if (!BarUiPianoCallError (&app, PIANO_REQUEST_LOGIN, &reqData, &pRet, &wRet, &error))
        {
        	*_error = [NSString stringWithCString:error encoding:NSASCIIStringEncoding];
        	return self;
        }
        
        
        //check to make sure we can get stations
        
        PianoReturn_t pRet2;
        WaitressReturn_t wRet2;
        
        if (!BarUiPianoCallError (&app, PIANO_REQUEST_GET_STATIONS, NULL, &pRet2, &wRet2, &error))
        {
        	*_error = [NSString stringWithCString:error encoding:NSASCIIStringEncoding];
        	return self;
        }
        
    }
    
    return self;
}

//level maybe from 0 to 100
-(void)setVolume:(int)level
{
	// -35 to 10
	
	app.settings.volume = level;
	NSLog(@"%d", app.settings.volume);
	app.player.scale = BarPlayerCalcScale (app.player.gain + app.settings.volume);
}

-(void)setDelegate:(id <PianobarDelegate>)anObject
{
	delegate = anObject;
}

-(BOOL)playing
{
	return playing;
}

-(void)nextSong
{
	app.player.doQuit = 1;
	pthread_mutex_unlock (&(&app.player)->pauseMutex);
	
	playing = YES;
}

-(void)playpause
{
	playing = !playing;
	/* already locked => unlock/unpause */
	if (pthread_mutex_trylock (&(&app.player)->pauseMutex) == EBUSY) {
		pthread_mutex_unlock (&(&app.player)->pauseMutex);
	}
}

-(BOOL)loveSong
{	
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	
	/* shared stations must be transformed */
	if (!(app.curStation)->isCreator) {
		//BarUiMsg (MSG_INFO, "Transforming station... ");
		const char* error;
		if (!BarUiPianoCallError (&app, PIANO_REQUEST_TRANSFORM_STATION, app.curStation, &pRet, &wRet, &error)) {
			NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
			[alert runModal];
			return NO;
		}
	}
	
	PianoReturn_t pRet2;
	WaitressReturn_t wRet2;
	
	PianoRequestDataRateSong_t reqData;
	reqData.song = app.playlist;
	reqData.rating = PIANO_RATE_LOVE;
	
	//BarUiMsg (MSG_INFO, "Loving song... ");
	const char* error;
	if(BarUiPianoCallError(&app, PIANO_REQUEST_RATE_SONG, &reqData, &pRet2, &wRet2, &error))
		return YES;
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}
	//BarUiActDefaultEventcmd ("songlove");
}

-(NSArray*)searchForMusicQuery:(NSString*)query isArtist:(BOOL)_artist
{	
	NSMutableArray* results = [NSMutableArray array];
	
	
	
	PianoSearchResult_t searchResult;
	
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataSearch_t reqData;
	
	reqData.searchStr = (char*)[query cStringUsingEncoding:NSUTF8StringEncoding];
	
	//BarUiMsg (MSG_INFO, "Searching... ");
	const char* error;
	if (!BarUiPianoCallError (&app, PIANO_REQUEST_SEARCH, &reqData, &pRet, &wRet, &error)) {
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return [NSArray array];
	}
	memcpy (&searchResult, &reqData.searchResult, sizeof (searchResult));

		//BarUiMsg (MSG_NONE, "\r");
	
	if (searchResult.songs != NULL || searchResult.artists != NULL)
	{
		/* songs or artists found */
		//BarUiMsg (MSG_QUESTION, "Is this an [a]rtist or [t]rack name? ");
		//BarReadline (selectBuf, sizeof (selectBuf), "at", 1, 0, curFd);
		
		if (_artist && searchResult.artists != NULL) {
			PianoArtist_t *startArtist = searchResult.artists;
			PianoArtist_t *tmpArtist = NULL;
			
			/* print all artists */
			tmpArtist = startArtist;
			while (tmpArtist != NULL) {
				//BarUiMsg (MSG_LIST, "%2u) %s\n", i, tmpArtist->name);
				[results addObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithCString:tmpArtist->musicId encoding:NSASCIIStringEncoding], @"id", 
									[NSString stringWithCString:tmpArtist->name encoding:NSUTF8StringEncoding], @"name", nil]];
				tmpArtist = tmpArtist->next;
			}
		} else if(!_artist && searchResult.songs != NULL) {
			PianoSong_t *startSong = searchResult.songs;
			
			PianoSong_t *tmpSong = NULL;
			
			/* print all songs */
			tmpSong = startSong;
			while (tmpSong != NULL) {
				//BarUiMsg (MSG_LIST, "%2u) %s - %s\n", i, tmpSong->artist, tmpSong->title);
				[results addObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithCString:tmpSong->musicId encoding:NSASCIIStringEncoding], @"id", 
									[NSString stringWithCString:tmpSong->artist encoding:NSUTF8StringEncoding], @"artist", 
									[NSString stringWithCString:tmpSong->title encoding:NSUTF8StringEncoding], @"name", nil]];
				tmpSong = tmpSong->next;
			}
		}
	} else {
        NSLog(@"Nothing found for search");
    }
	
	PianoDestroySearchResult (&searchResult);
	
	return [NSArray arrayWithArray:results];
}

-(NSArray*)getSuggestions
{
	NSMutableArray* results = [NSMutableArray array];
	/*
	char *similarToId = app.playlist->musicId;
	
	PianoSearchResult_t searchResult;

	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataGetSeedSuggestions_t reqData;
	
	reqData.musicId = similarToId;
	reqData.max = 20;
	
	//BarUiMsg (MSG_INFO, "Receiving suggestions... ");
	const char* error;
	if (!BarUiPianoCallError (&app, PIANO_REQUEST_GET_SEED_SUGGESTIONS, &reqData, &pRet, &wRet, &error))
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return [NSArray array];
	}
	memcpy (&searchResult, &reqData.searchResult, sizeof (searchResult));
	
	
	PianoArtist_t *startArtist = searchResult.artists;
	PianoArtist_t *tmpArtist = NULL;
	

	tmpArtist = startArtist;
	while (tmpArtist != NULL) {
		//BarUiMsg (MSG_LIST, "%2u) %s\n", i, tmpArtist->name);
		[results addObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithCString:tmpArtist->musicId encoding:NSASCIIStringEncoding], @"id", 
							[NSString stringWithCString:tmpArtist->name encoding:NSUTF8StringEncoding], @"name", nil]];
		tmpArtist = tmpArtist->next;
	}
	
	PianoDestroySearchResult (&searchResult);*/
	
	return [NSArray arrayWithArray:results];
}

-(NSString*)getSongExplanation
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataExplain_t reqData;
	
	if (app.curStation == NULL || app.playlist == NULL)
		return @"";
	
	reqData.song = app.playlist;
	
	NSString* s = @"";
	
	//BarUiMsg (MSG_INFO, "Receiving explanation... ");
	//if (BarUiActDefaultPianoCall (PIANO_REQUEST_EXPLAIN, &reqData)) {
	const char* error;
	if(BarUiPianoCallError (&app, PIANO_REQUEST_EXPLAIN, &reqData, &pRet, &wRet, &error))
	{
		//BarUiMsg (MSG_INFO, "%s\n", reqData.retExplain);
		s = [NSString stringWithCString:reqData.retExplain encoding:NSUTF8StringEncoding];
		free (reqData.retExplain);
	}
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
	}

	
	//BarUiActDefaultEventcmd ("songexplain");
	
	return s;
}

-(BOOL)bookmarkCurrentArtist
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	//BarUiMsg (MSG_INFO, "Bookmarking artist... ");
	//BarUiActDefaultPianoCall (PIANO_REQUEST_BOOKMARK_ARTIST, app->playlist);
	const char* error;
	if(BarUiPianoCallError (&app, PIANO_REQUEST_BOOKMARK_ARTIST, app.playlist, &pRet, &wRet, &error))
		return YES;
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}

	//BarUiActDefaultEventcmd ("artistbookmark");
}

-(BOOL)bookmarkCurrentSong
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	//BarUiMsg (MSG_INFO, "Bookmarking song... ");
	//BarUiActDefaultPianoCall (PIANO_REQUEST_BOOKMARK_SONG, app->playlist);
	const char* error;
	if(BarUiPianoCallError (&app, PIANO_REQUEST_BOOKMARK_SONG, app.playlist, &pRet, &wRet, &error))
		return YES;
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}
	//BarUiActDefaultEventcmd ("songbookmark");
}

-(NSArray*)upcomingSongs
{
	if (app.curStation == NULL || app.playlist == NULL)
		return [NSArray array];
	
	NSMutableArray* upcomingSongs = [NSMutableArray array];
	
	PianoSong_t *nextSong = app.playlist->next;
	while (nextSong != NULL) {
		//BarUiMsg (MSG_LIST, "%2i) \"%s\" by \"%s\"\n", i, nextSong->title, nextSong->artist);
		[upcomingSongs addObject:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithCString:nextSong->title encoding:NSUTF8StringEncoding], @"title",
								  [NSString stringWithCString:nextSong->artist encoding:NSUTF8StringEncoding], @"artist", nil]];
		nextSong = nextSong->next;
	}
	
	return [NSArray arrayWithArray:upcomingSongs];
}

-(BOOL)addSharedStationById:(NSString*)_id
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataCreateStation_t reqData;
	
	//BarUiMsg (MSG_QUESTION, "Station id: ");

	reqData.token = (char*)[_id cStringUsingEncoding:NSASCIIStringEncoding];
	reqData.type = PIANO_MUSICTYPE_INVALID;
	//BarUiMsg (MSG_INFO, "Adding shared station... ");
	//BarUiActDefaultPianoCall (PIANO_REQUEST_CREATE_STATION, &reqData);
	const char* error;
	if(BarUiPianoCallError(&app, PIANO_REQUEST_CREATE_STATION, &reqData, &pRet, &wRet, &error))
		return YES;
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}
	//BarUiActDefaultEventcmd ("stationaddshared");
	
}

-(BOOL)deleteStationWithId:(NSString*)_id
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	//BarUiMsg (MSG_QUESTION, "Really delete \"%s\"? [yN] ", app->curStation->name);
	
	//PianoStation_t **sortedStations = NULL;
	//size_t stationCount, i;
	
	PianoStation_t *td;
	
	PianoHandle_t *ph = &app.ph;
	
	/*
	sortedStations = BarSortedStations (ph->stations, &stationCount, app.settings.sortOrder);
	for (i = 0; i < stationCount; i++)
	{
		const PianoStation_t *currStation = sortedStations[i];
		if(strcmp(currStation->id, [_id cStringUsingEncoding:NSASCIIStringEncoding]) == 0)
		{
			td = sortedStations[i];
			break;
		}
	}*/
	
	td = PianoFindStationById(ph->stations, [_id cStringUsingEncoding:NSASCIIStringEncoding]);
	
	BOOL ics = (app.curStation == td);

	//BarUiMsg (MSG_INFO, "Deleting station... ");
	//if (BarUiActDefaultPianoCall (PIANO_REQUEST_DELETE_STATION, app->curStation)) {
	const char* error;
	if(BarUiPianoCallError(&app, PIANO_REQUEST_DELETE_STATION, td, &pRet, &wRet, &error))
	{
		if(ics)
		{
			[self nextSong];
			PianoDestroyPlaylist (app.playlist);
			app.playlist = NULL;
			app.curStation = NULL;
		}
		
		return YES;
	}
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}
	//BarUiActDefaultEventcmd ("stationdelete");
}

-(BOOL)createStationById:(NSString*)_id
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataCreateStation_t reqData;
	
	reqData.token = (char*)[_id cStringUsingEncoding:NSASCIIStringEncoding];
	
	if (reqData.token != NULL) {
		reqData.type = PIANO_MUSICTYPE_INVALID;
		//BarUiMsg (MSG_INFO, "Creating station... ");
		//BarUiActDefaultPianoCall (PIANO_REQUEST_CREATE_STATION, &reqData);
		
		const char* error;
		if(BarUiPianoCallError(&app, PIANO_REQUEST_CREATE_STATION, &reqData, &pRet, &wRet, &error))
			return YES;
		else
		{
			NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
			[alert runModal];
			return NO;
		}
		//free (reqData.id);
		//BarUiActDefaultEventcmd ("stationcreate");
	}
	
	return NO;
}

-(BOOL)addMusicToCurrentStationById:(NSString*)_id
{
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	PianoRequestDataAddSeed_t reqData;
	
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	reqData.musicId = (char*)[_id cStringUsingEncoding:NSASCIIStringEncoding];
	
	if (reqData.musicId != NULL) {
		
		/* shared stations must be transformed */
		if (!(app.curStation)->isCreator) {
			//BarUiMsg (MSG_INFO, "Transforming station... ");
			const char* error;
			if (!BarUiPianoCallError(&app, PIANO_REQUEST_TRANSFORM_STATION, app.curStation, &pRet, &wRet, &error))
			{
				NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
				[alert runModal];
				return NO;
			}
		}
		
		PianoReturn_t pRet2;
		WaitressReturn_t wRet2;
		
		reqData.station = app.curStation;
		
		//BarUiMsg (MSG_INFO, "Adding music to station... ");
		//BarUiActDefaultPianoCall (PIANO_REQUEST_ADD_SEED, &reqData);
		const char* error;
		if(BarUiPianoCallError(&app, PIANO_REQUEST_ADD_SEED, &reqData, &pRet2, &wRet2, &error))
		{
			free (reqData.musicId);
			return YES;
		}
		else
		{
			free (reqData.musicId);
			NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
			[alert runModal];
			return NO;
		}
		
		
		
		//BarUiActDefaultEventcmd ("stationaddmusic");
	}
	else
		return NO;
	
	return YES;
}

-(BOOL)banSong
{
	if (app.curStation == NULL || app.playlist == NULL)
		return NO;
	
	PianoReturn_t pRet;
	WaitressReturn_t wRet;
	
	/* shared stations must be transformed */
	if (!(app.curStation)->isCreator) {
		//BarUiMsg (MSG_INFO, "Transforming station... ");
		const char* error;
		if (!BarUiPianoCallError(&app, PIANO_REQUEST_TRANSFORM_STATION, app.curStation, &pRet, &wRet, &error))
		{
			NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
			[alert runModal];
			return NO;
		}
	}
	
	PianoReturn_t pRet2;
	WaitressReturn_t wRet2;
	
	PianoRequestDataRateSong_t reqData;
	reqData.song = app.playlist;
	reqData.rating = PIANO_RATE_BAN;
	
	//BarUiMsg (MSG_INFO, "Banning song... ");
	const char* error;
	if (BarUiPianoCallError(&app, PIANO_REQUEST_RATE_SONG, &reqData, &pRet2, &wRet2, &error)) {
		[self nextSong];
	}
	else
	{
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
		[alert runModal];
		return NO;
	}
	//BarUiActDefaultEventcmd ("songban");
	
	return YES;
}

-(void)selectStationById:(NSString*)_id
{
	if(app.curStation == NULL || strcmp(app.curStation->id, [_id cStringUsingEncoding:NSASCIIStringEncoding]) != 0)
	{
		PianoStation_t **sortedStations = NULL;
		//size_t stationCount, i;
		
		PianoHandle_t *ph = &app.ph;
		
		/*
		sortedStations = BarSortedStations (ph->stations, &stationCount, app.settings.sortOrder);
		for (i = 0; i < stationCount; i++)
		{
			const PianoStation_t *currStation = sortedStations[i];
			if(strcmp(currStation->id, [_id cStringUsingEncoding:NSASCIIStringEncoding]) == 0)
			{
				app.curStation = sortedStations[i];
				break;
			}
		}*/
		
		app.curStation = PianoFindStationById(ph->stations, [_id cStringUsingEncoding:NSASCIIStringEncoding]);
		
		if(app.playlist != NULL)
		{
			(&app.player)->doQuit = 1;
			pthread_mutex_unlock (&(&app.player)->pauseMutex);
			PianoDestroyPlaylist (app.playlist);
			app.playlist = NULL;
		}
		
		playing = YES;
		
		free (sortedStations);
	}
}


-(NSArray*)getStations
{	
	NSMutableArray* stations = [NSMutableArray array];
	
	PianoStation_t **sortedStations = NULL;
	size_t stationCount, i;
	
	PianoHandle_t *ph = &app.ph;
	
    /*
     static PianoStation_t **BarSortedStations (PianoStation_t *unsortedStations,
     size_t *retStationCount, BarStationSorting_t order) {
     */
    
	/* sort and print stations */
	sortedStations = BarSortedStations (ph->stations, &stationCount, app.settings.sortOrder);
	for (i = 0; i < stationCount; i++)
	{
		const PianoStation_t *currStation = sortedStations[i];
		NSMutableDictionary* dict = [NSMutableDictionary dictionary];
		
		[dict setObject:[NSString stringWithCString:currStation->id encoding:NSASCIIStringEncoding] forKey:@"id"];
		[dict setObject:[NSString stringWithCString:currStation->name encoding:NSUTF8StringEncoding] forKey:@"name"];
		[dict setObject:[NSNumber numberWithBool:currStation->useQuickMix] forKey:@"useQuickMix"];
		[dict setObject:[NSNumber numberWithBool:currStation->isQuickMix] forKey:@"isQuickMix"];
		[dict setObject:[NSNumber numberWithBool:currStation->isCreator] forKey:@"isCreator"];
		
		BOOL contains = NO;
		
		for(NSDictionary* t in stations)
		{
			if([[t objectForKey:@"id"] isEqualToString:[dict objectForKey:@"id"]])
				contains = YES;
		}
		
		if(!contains)
			[stations addObject:dict];
	}
	
	free (sortedStations);
	
	return [NSArray arrayWithArray:stations];
}

-(void)startStreaming
{
	playing = YES;
	[self performSelectorInBackground:@selector(play) withObject:nil];
}

-(void)stopStreaming
{
//	app.doQuit = 1;
//	[self nextSong];
//	
//	//[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:2]];
//	
//	if (app.player.mode != PLAYER_FREED) {
//		pthread_join (playerThread, NULL);
//	}
//	if (ctlFd != NULL) {
//		fclose (ctlFd);
//	}
//	PianoDestroy (&app.ph);
//	PianoDestroyPlaylist (app.songHistory);
//	PianoDestroyPlaylist (app.playlist);
//	WaitressFree (&app.waith);
//	ao_shutdown();
//	gnutls_global_deinit ();
//	BarSettingsDestroy (&app.settings);
    
    [self nextSong];
    
    if (app.input.fds[1] != -1) {
        close (app.input.fds[1]);
    }
    
    PianoDestroy (&app.ph);
    PianoDestroyPlaylist (app.songHistory);
    PianoDestroyPlaylist (app.playlist);
    WaitressFree (&app.waith);
    ao_shutdown();
    gnutls_global_deinit ();
    BarSettingsDestroy (&app.settings);
}

-(void)dealloc
{
	[super dealloc];
}

-(void)play
{
    /*
     
     void BarUiStartEventCmd (const BarSettings_t *settings, const char *type,
     const PianoStation_t *curStation, const PianoSong_t *curSong,
     const struct audioPlayer *player, PianoStation_t *stations,
     PianoReturn_t pRet, WaitressReturn_t wRet) {
     
     */
    
    
	
//	/* little hack, needed to signal: hey! we need a playlist, but don't
//	 * free anything (there is nothing to be freed yet) */
//	memset (&app.player, 0, sizeof (app.player));
//	
//	while (!app.doQuit) {
//		NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
//		/* song finished playing, clean up things/scrobble song */
//		if (app.player.mode == PLAYER_FINISHED_PLAYBACK) {
//			BarUiStartEventCmd (&app.settings, "songfinish", app.curStation,
//								app.playlist, &app.player, NULL, PIANO_RET_OK, WAITRESS_RET_OK);
//			/* FIXME: pthread_join blocks everything if network connection
//			 * is hung up e.g. */
//			void *threadRet;
//			pthread_join (playerThread, &threadRet);
//			/* don't continue playback if thread reports error */
//			if (threadRet != (void *) PLAYER_RET_OK) {
//				app.curStation = NULL;
//			}
//			memset (&app.player, 0, sizeof (app.player));
//		}
//		
//		/* check whether player finished playing and start playing new
//		 * song */
//		if (app.player.mode >= PLAYER_FINISHED_PLAYBACK ||
//			app.player.mode == PLAYER_FREED) {
//			if (app.curStation != NULL) {
//				/* what's next? */
//				if (app.playlist != NULL) {
//					if (app.settings.history != 0) {
//						/* prepend song to history list */
//						PianoSong_t *tmpSong = app.songHistory;
//						app.songHistory = app.playlist;
//						/* select next song */
//						app.playlist = app.playlist->next;
//						app.songHistory->next = tmpSong;
//						
//						/* limit history's length */
//						/* start with 1, so we're stopping at n-1 and have the
//						 * chance to set ->next = NULL */
//						unsigned int i = 1;
//						tmpSong = app.songHistory;
//						while (i < app.settings.history && tmpSong != NULL) {
//							tmpSong = tmpSong->next;
//							++i;
//						}
//						/* if too many songs in history... */
//						if (tmpSong != NULL) {
//							PianoSong_t *delSong = tmpSong->next;
//							tmpSong->next = NULL;
//							if (delSong != NULL) {
//								PianoDestroyPlaylist (delSong);
//							}
//						}
//					} else {
//						/* don't keep history */
//						app.playlist = app.playlist->next;
//					}
//				}
//				if (app.playlist == NULL) {
//					PianoReturn_t pRet;
//					WaitressReturn_t wRet;
//					PianoRequestDataGetPlaylist_t reqData;
//					reqData.station = app.curStation;
//					reqData.format = app.settings.audioFormat;
//					
//					//BarUiMsg (MSG_INFO, "Receiving new playlist... ");
//					const char* error;
//					
//					
//					//Relogin after 5 hour limit (hopefully)
//					BOOL res_t = BarUiPianoCallError(&app, PIANO_REQUEST_GET_PLAYLIST, &reqData, &pRet, &wRet, &error);
//					if(!res_t && [[NSString stringWithCString:error encoding:NSASCIIStringEncoding] isEqual:@"Invalid auth token."])
//					{
//						[delegate connectionDidExpire];
//					}
//					
//					if (!res_t)
//					{
//						//NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:[NSString stringWithCString:error encoding:NSASCIIStringEncoding]];
//						//[alert runModal];
//						app.curStation = NULL;
//					} else {
//						app.playlist = reqData.retPlaylist;
//						if (app.playlist == NULL) {
//							//BarUiMsg (MSG_INFO, "No tracks left.\n");
//							app.curStation = NULL;
//						}
//					}
//					BarUiStartEventCmd (&app.settings, "stationfetchplaylist",
//										app.curStation, app.playlist, &app.player, NULL, pRet,
//										wRet);
//				}
//				/* song ready to play */
//				if (app.playlist != NULL) {
//					
//					//Song Start Callback
//					PianoSong_t *song = app.playlist;
//					//PianoStation_t *station = (app.curStation->isQuickMix ? PianoFindStationById (app.ph.stations, app.playlist->stationId) : NULL);
//					
//					NSMutableDictionary* songInfo = [NSMutableDictionary dictionary];
//					[songInfo setObject:[NSString stringWithCString:song->title encoding:NSUTF8StringEncoding] forKey:@"title"];
//					[songInfo setObject:[NSString stringWithCString:song->artist encoding:NSUTF8StringEncoding] forKey:@"artist"];
//					[songInfo setObject:[NSString stringWithCString:song->album encoding:NSUTF8StringEncoding] forKey:@"album"];
//					[songInfo setObject:[NSString stringWithCString:app.curStation->name encoding:NSUTF8StringEncoding] forKey:@"station"];
//					[songInfo setObject:[NSNumber numberWithBool:(song->rating == PIANO_RATE_LOVE)] forKey:@"loveit"];
//					NSImage* albumArt = [[[NSImage alloc] initWithContentsOfURL:[NSURL URLWithString:[NSString stringWithCString:song->coverArt encoding:NSASCIIStringEncoding]]] autorelease];
//					if(albumArt != nil)
//						[songInfo setObject:albumArt forKey:@"albumArt"];
//					else
//						[songInfo setObject:[NSImage imageNamed:@"not_found"] forKey:@"albumArt"];
//					
//					[delegate songDidStart:songInfo];
//					
//					if (app.playlist->audioUrl == NULL) {
//						//BarUiMsg (MSG_ERR, "Invalid song url.\n");
//						NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:@"Invalid song url"];
//						[alert runModal];
//					} else {
//						/* setup player */
//						memset (&app.player, 0, sizeof (app.player));
//						
//						WaitressInit (&app.player.waith);
//						WaitressSetUrl (&app.player.waith, app.playlist->audioUrl);
//						
//						/* set up global proxy, player is NULLed on songfinish */
//						if (app.settings.proxy != NULL) {
//							WaitressSetProxy (&app.player.waith, app.settings.proxy);
//						}
//						
//						app.player.gain = app.playlist->fileGain;
//						app.player.scale = BarPlayerCalcScale (app.player.gain + app.settings.volume);
//						app.player.audioFormat = app.playlist->audioFormat;
//						
//						/* throw event */
//						BarUiStartEventCmd (&app.settings, "songstart",
//											app.curStation, app.playlist, &app.player,
//											NULL, PIANO_RET_OK, WAITRESS_RET_OK);
//						
//						/* prevent race condition, mode must _not_ be FREED if
//						 * thread has been started */
//						app.player.mode = PLAYER_STARTING;
//						/* start player */
//						pthread_create (&playerThread, NULL, BarPlayerThread,
//										&app.player);
//					} /* end if audioUrl == NULL */
//				} /* end if playlist != NULL */
//			} /* end if curStation != NULL */
//		}
//		
//		/* select modifies its arguments => copy the set */
//		memcpy (&readSetCopy, &readSet, sizeof (readSet));
//		selectTimeout.tv_sec = 1;
//		selectTimeout.tv_usec = 0;
//		
//		
//		
//		/* in the meantime: wait for user actions */
//		/*if (select (maxFd, &readSetCopy, NULL, NULL, &selectTimeout) > 0) {
//			FILE *curFd = NULL;
//			
//			if (FD_ISSET(selectFds[0], &readSetCopy)) {
//				curFd = stdin;
//			} else if (FD_ISSET(selectFds[1], &readSetCopy)) {
//				curFd = ctlFd;
//			}
//			buf = fgetc (curFd);
//			
//			size_t i;
//			for (i = 0; i < BAR_KS_COUNT; i++) {
//				if (app.settings.keys[i] == buf) {
//					static const BarKeyShortcutFunc_t idToF[] = {BarUiActHelp,
//						BarUiActLoveSong, BarUiActBanSong,
//						BarUiActAddMusic, BarUiActCreateStation,
//						BarUiActDeleteStation, BarUiActExplain,
//						BarUiActStationFromGenre, BarUiActHistory,
//						BarUiActSongInfo, BarUiActAddSharedStation,
//						BarUiActMoveSong, BarUiActSkipSong, BarUiActPause,
//						BarUiActQuit, BarUiActRenameStation,
//						BarUiActSelectStation, BarUiActTempBanSong,
//						BarUiActPrintUpcoming, BarUiActSelectQuickMix,
//						BarUiActDebug, BarUiActBookmark};
//					idToF[i] (&app, curFd);
//					break;
//				}
//			}
//		}*/
//		
//		
//		//Time callback
//		if (app.player.mode >= PLAYER_SAMPLESIZE_INITIALIZED &&
//			app.player.mode < PLAYER_FINISHED_PLAYBACK)
//		{
//			[delegate positionUpdate:(int)(app.player.songPlayed / BAR_PLAYER_MS_TO_S_FACTOR) duration:(int)(app.player.songDuration / BAR_PLAYER_MS_TO_S_FACTOR)];
//		}
//		
//		
//		/* show time */
//		/*
//		if (app.player.mode >= PLAYER_SAMPLESIZE_INITIALIZED &&
//			app.player.mode < PLAYER_FINISHED_PLAYBACK) {
//			// Ugly: songDuration is unsigned _long_ int! Lets hope this won't
//			 // overflow 
//			int songRemaining = (signed long int) (app.player.songDuration -
//												   app.player.songPlayed) / BAR_PLAYER_MS_TO_S_FACTOR;
//			char pos = 0;
//			if (songRemaining < 0) {
//				// Use plus sign if song is longer than expected 
//				pos = 1;
//				songRemaining = -songRemaining;
//			}
//			BarUiMsg (MSG_TIME, "%c%02i:%02i/%02i:%02i\r", (pos ? '+' : '-'),
//					  songRemaining / 60, songRemaining % 60,
//					  app.player.songDuration / BAR_PLAYER_MS_TO_S_FACTOR / 60,
//					  app.player.songDuration / BAR_PLAYER_MS_TO_S_FACTOR % 60);
//		}
//		*/
//		
//		//without this sleep, the loop causes the cpu usage to soar to 100%
//		[NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];
//		[pool release];
//	}
	
	/* 
	if (app.player.mode != PLAYER_FREED) {
		pthread_join (playerThread, NULL);
	}
	if (ctlFd != NULL) {
		fclose (ctlFd);
	}
	PianoDestroy (&app.ph);
	PianoDestroyPlaylist (app.songHistory);
	PianoDestroyPlaylist (app.playlist);
	ao_shutdown();
	BarSettingsDestroy (&app.settings);*/
	
//	BarMainGetLoginCredentials (&app.settings, &app.input);
    
//	if (!BarMainGetStations (app)) {
//		return;
//	}
    
//	BarMainGetInitialStation (app);
    
    pthread_t playerThread;
    
	/* little hack, needed to signal: hey! we need a playlist, but don't
	 * free anything (there is nothing to be freed yet) */
	memset (&app.player, 0, sizeof (app.player));
    
	while (!app.doQuit) {
		/* song finished playing, clean up things/scrobble song */
		if (app.player.mode == PLAYER_FINISHED_PLAYBACK) {
			BarMainPlayerCleanup (&app, &playerThread);
		}
        
		/* check whether player finished playing and start playing new
		 * song */
		if (app.player.mode >= PLAYER_FINISHED_PLAYBACK ||
            app.player.mode == PLAYER_FREED) {
			if (app.curStation != NULL) {
				/* what's next? */
				if (app.playlist != NULL) {

					//PianoSong_t *histsong = app.playlist;
					app.playlist = app.playlist->next;
					//BarUiHistoryPrepend(&app, histsong);
				}
				if (app.playlist == NULL) {

					BarMainGetPlaylist (&app);
				}
				/* song ready to play */
				if (app.playlist != NULL) {
                    
                    
                    //Song Start Callback
                   	PianoSong_t *song = app.playlist;

                   	NSMutableDictionary* songInfo = [NSMutableDictionary dictionary];
                   	[songInfo setObject:[NSString stringWithCString:song->title encoding:NSUTF8StringEncoding] forKey:@"title"];
                   	[songInfo setObject:[NSString stringWithCString:song->artist encoding:NSUTF8StringEncoding] forKey:@"artist"];
                   	[songInfo setObject:[NSString stringWithCString:song->album encoding:NSUTF8StringEncoding] forKey:@"album"];
                   	[songInfo setObject:[NSString stringWithCString:app.curStation->name encoding:NSUTF8StringEncoding] forKey:@"station"];
                   	[songInfo setObject:[NSNumber numberWithBool:(song->rating == PIANO_RATE_LOVE)] forKey:@"loveit"];
                   	NSImage* albumArt = [[[NSImage alloc] initWithContentsOfURL:[NSURL URLWithString:[NSString stringWithCString:song->coverArt encoding:NSASCIIStringEncoding]]] autorelease];
                   	if(albumArt != nil)
                   		[songInfo setObject:albumArt forKey:@"albumArt"];
                   	else
                   		[songInfo setObject:[NSImage imageNamed:@"not_found"] forKey:@"albumArt"];
                   	[delegate songDidStart:songInfo];
                    
                    
                    
					BarMainStartPlayback (&app, &playerThread);
				}
			}
		}
        
       //Time callback
       if (app.player.mode >= PLAYER_SAMPLESIZE_INITIALIZED &&
       	app.player.mode < PLAYER_FINISHED_PLAYBACK)
       {
       	[delegate positionUpdate:(int)(app.player.songPlayed / BAR_PLAYER_MS_TO_S_FACTOR) duration:(int)(app.player.songDuration / BAR_PLAYER_MS_TO_S_FACTOR)];
       }
        
        
//		BarMainHandleUserInput (app);
        
		/* show time */
//		if (app.player.mode >= PLAYER_SAMPLESIZE_INITIALIZED &&
//            app.player.mode < PLAYER_FINISHED_PLAYBACK) {
//			BarMainPrintTime (app);
//		}
        [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:1.0]];
    }
    
//	if (app.player.mode != PLAYER_FREED) {
//		pthread_join (playerThread, NULL);
//	}
    
}

@end
