//
//  PianopubAppDelegate.m
//  Pianopub
//
//  Created by Remi Bernotavicius on 8/24/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "PianopubAppDelegate.h"
#import "PSearchField.h"
#import "EMKeychainItem.h"
#import "PWindow.h"

@implementation PianopubAppDelegate

#pragma mark Init / Dealloc

-(id)init
{
	if((self = [super init]))
	{
		songInformation = [[NSMutableDictionary alloc] init];
		searchResults = [[NSMutableArray alloc] init];
		loginInformation = [[NSMutableDictionary alloc] init];
		trackInformation = [[NSMutableDictionary alloc] initWithObjectsAndKeys:@"0:00", @"position", @"0:00", @"duration", [NSNumber numberWithBool:NO], @"playing", nil];
		stations = [[NSMutableArray alloc] init];
		
		[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
																 [NSNumber numberWithInt:1], @"notifications", 
																 [NSNumber numberWithInt:1], @"notifications_background", 
																 [NSNumber numberWithInt:1], @"appleremote", nil]];
	}
	
	return self;
}

-(void)dealloc
{
	[controlManager release];
	[loginInformation release];
	[_pianobar release];
	[songInformation release];
	[trackInformation release];
	[searchResults release];
	[stations release];
	[super dealloc];
}

#pragma mark Delegate Methods

-(void)awakeFromNib
{
	if([[NSUserDefaults standardUserDefaults] boolForKey:@"stationsHidden"])
	{
		[self hideStations:self];
	}
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
	controlManager = [[MediaControls alloc] init];
	
	[GrowlApplicationBridge setGrowlDelegate:self];
	[[NSApplication sharedApplication] setDelegate:self];
}

-(void)windowDidLoad:(NSWindow*)theWindow
{
	NSString* username = [[NSUserDefaults standardUserDefaults] objectForKey:@"username"];
	
	EMKeychainItem* keychainItem = nil;
	if(username != nil)
	{
		[self willChangeValueForKey:@"loginInformation"];
		[loginInformation setObject:username forKey:@"username"];
		[self didChangeValueForKey:@"loginInformation"];
		
		keychainItem = [EMGenericKeychainItem genericKeychainItemForService:@"pianopub" withUsername:username];
	}
	
	if(keychainItem == nil)
		[(PWindow*)loginWindow showAsSheet:self];
	else
	{
		[self willChangeValueForKey:@"loginInformation"];
		[loginInformation setObject:[keychainItem password] forKey:@"password"];
		//[loginInformation setObject:keychainItem forKey:@"keychainItem"];
		[self didChangeValueForKey:@"loginInformation"];
		[self login:self];
	}
}

-(BOOL)applicationShouldHandleReopen:(NSApplication *)theApplication hasVisibleWindows:(BOOL)flag
{
	if(!flag)
	{
		[window makeKeyAndOrderFront:self];
	}
	return YES;
}

-(void)tableViewDidDoubleClick:(id)sender
{
	[_pianobar selectStationById:[[[stationsController selectedObjects] objectAtIndex:0] objectForKey:@"id"]];
	
	[self updatePlaying];
}

#pragma mark Accessor Methods

-(NSMutableDictionary*)loginInformation
{
	return loginInformation;
}

-(NSMutableDictionary*)songInformation
{
	return songInformation;
}

-(NSMutableDictionary*)trackInformation
{
	return trackInformation;
}

-(NSMutableArray*)stations
{
	return stations;
}

-(NSMutableArray*)searchResults
{
	return searchResults;
}

#pragma mark Pianobar Delegate Methods

-(void)songDidStart:(NSDictionary*)songInfo
{	
	[self willChangeValueForKey:@"songInformation"];
	[songInformation addEntriesFromDictionary:songInfo];
	[self didChangeValueForKey:@"songInformation"];
	
	[self updatePlaying];
	
	if([[NSUserDefaults standardUserDefaults] boolForKey:@"notifications"])
	{
		if ([[NSUserDefaults standardUserDefaults] boolForKey:@"notifications_background"]) {
			if(![window isKeyWindow])
				[self sendNotification];
		}
		else
			[self sendNotification];
	}
}

-(void)positionUpdate:(int)position duration:(int)duration
{
	[self willChangeValueForKey:@"trackInformation"];
	[trackInformation setObject:[NSNumber numberWithDouble:(double)position / duration * 100] forKey:@"percentComplete"];
	NSString* min = position % 60 < 10 ? [NSString stringWithFormat:@"0%d", position % 60] : [NSString stringWithFormat:@"%d", position % 60];
	[trackInformation setObject:[NSString stringWithFormat:@"%d:%@", position / 60, min] forKey:@"position"];
	min = duration % 60 < 10 ? [NSString stringWithFormat:@"0%d", duration % 60] : [NSString stringWithFormat:@"%d", duration % 60];
	[trackInformation setObject:[NSString stringWithFormat:@"%d:%@", duration / 60, min] forKey:@"duration"];
	[self didChangeValueForKey:@"trackInformation"];
}

-(void)connectionDidExpire
{
	NSString* oldStation;
	int i;
	for(i=0;i<[stations count];i++)
	{
		if([[stations objectAtIndex:i] objectForKey:@"playing"] != nil)
		{
			oldStation = [[stations objectAtIndex:i] objectForKey:@"id"];
			break;
		}
	}
	[self clearInformation];
	[_pianobar stopStreaming];
	[self login:self];
	[_pianobar selectStationById:oldStation];
}

#pragma mark Internal Methods

-(void)sendNotification
{
	[GrowlApplicationBridge notifyWithTitle:[songInformation objectForKey:@"title"] 
								description:[NSString stringWithFormat:@"%@\n%@", [songInformation objectForKey:@"artist"], [songInformation objectForKey:@"album"]] 
						   notificationName:@"Song Changed" 
								   iconData:[[songInformation objectForKey:@"albumArt"] TIFFRepresentation]
								   priority:0 
								   isSticky:NO clickContext:nil];
}

-(void)updateStations
{
	[self willChangeValueForKey:@"stations"];
	[stations removeAllObjects];
	[stations addObjectsFromArray:[_pianobar getStations]];
	[self didChangeValueForKey:@"stations"];
	[self updatePlaying];
}

-(void)clearInformation
{
	[self willChangeValueForKey:@"songInformation"];
	[songInformation removeAllObjects];
	[self didChangeValueForKey:@"songInformation"];
	
	[self willChangeValueForKey:@"trackInformation"];
	[trackInformation setObject:@"0:00" forKey:@"position"];
	[trackInformation setObject:@"0:00" forKey:@"duration"];
	[trackInformation setObject:[NSNumber numberWithInt:0] forKey:@"percentComplete"];
	[self didChangeValueForKey:@"trackInformation"];
}

-(void)updatePlaying
{	
	[self willChangeValueForKey:@"trackInformation"];
	[trackInformation setObject:[NSNumber numberWithBool:[_pianobar playing]] forKey:@"playing"];
	[self didChangeValueForKey:@"trackInformation"];
	
	[self willChangeValueForKey:@"stations"];
	int i;
	for(i=0;i < [stations count]; i++)
	{
		if([[[stations objectAtIndex:i] objectForKey:@"name"] isEqual:[songInformation objectForKey:@"station"]])
			[[stations objectAtIndex:i] setObject:[NSImage imageNamed:@"NSGoRightTemplate"] forKey:@"playing"];
		else
			[[stations objectAtIndex:i] removeObjectForKey:@"playing"];
	}
	[self didChangeValueForKey:@"stations"];
}

#pragma mark Interface Actions

-(IBAction)volumeChange:(id)sender
{
	[_pianobar setVolume:[sender intValue]];
}

-(void)volumeUp
{
	[volumeSlider setIntValue:[volumeSlider intValue] + 1];
	[_pianobar setVolume:[volumeSlider intValue]];
}

-(void)volumeDown
{
	[volumeSlider setIntValue:[volumeSlider intValue] - 1];
	[_pianobar setVolume:[volumeSlider intValue]];
}

-(IBAction)bookmarkSong:(id)sender
{
    NSAlert* alert = [[[NSAlert alloc] init] autorelease];
	[alert addButtonWithTitle:@"OK"];
    [alert setIcon:[NSImage imageNamed:@"bookmark"]];
    if([_pianobar bookmarkCurrentSong])
        [alert setMessageText:[NSString stringWithFormat:@"Song \"%@\" Bookmarked", [songInformation objectForKey:@"title"]]];
    else
        [alert setMessageText:[NSString stringWithFormat:@"There was a problem bookmarking Song \"%@\"", [songInformation objectForKey:@"title"]]];
	[alert runModal];
}

-(IBAction)bookmarkArtist:(id)sender
{
    NSAlert* alert = [[[NSAlert alloc] init] autorelease];
	[alert addButtonWithTitle:@"OK"];
    [alert setIcon:[NSImage imageNamed:@"bookmark"]];
    if([_pianobar bookmarkCurrentArtist])
        [alert setMessageText:[NSString stringWithFormat:@"Artist \"%@\" Bookmarked", [songInformation objectForKey:@"artist"]]];
    else
        [alert setMessageText:[NSString stringWithFormat:@"There was a Problem Bookmarking Artist \"%@\"", [songInformation objectForKey:@"artist"]]];
	[alert runModal];
}

-(IBAction)login:(id)sender
{
	if(_pianobar != nil)
	{
		[_pianobar release];
	}
	
	NSString* error = nil;
	NSString* username = [loginInformation objectForKey:@"username"];
	NSString* password = [loginInformation objectForKey:@"password"];
	username = (username == nil) ? @"" : username;
	password = (password == nil) ? @"" : password;
    
    PianoAudioFormat_t quality = [[NSUserDefaults standardUserDefaults] integerForKey:@"quality"];
	
	_pianobar = [[Pianobar alloc] initWithUsername:username andPassword:password quality:quality error:&error];
	[_pianobar setDelegate:self];
	
	if(error != nil)
	{
		NSLog(@"%@", error);
		NSAlert* alert = [NSAlert alertWithMessageText:@"Error" defaultButton:@"OK" alternateButton:nil otherButton:nil informativeTextWithFormat:error];
		[alert runModal];
		if(sender == self)
		{
			[(PWindow*)loginWindow showAsSheet:self];
		}
	}
	else
	{
		//save password if not from self
		if(sender != self)
		{
			EMKeychainItem* keychainItem = [EMGenericKeychainItem genericKeychainItemForService:@"pianopub" withUsername:[loginInformation objectForKey:@"username"]];
			if(keychainItem != nil)
			{
				[keychainItem setUsername:[loginInformation objectForKey:@"username"]];
				[keychainItem setPassword:[loginInformation objectForKey:@"password"]];
			}
			else
			{
				keychainItem = [EMGenericKeychainItem addGenericKeychainItemForService:@"pianopub" withUsername:[loginInformation objectForKey:@"username"] password:[loginInformation objectForKey:@"password"]];
				[loginInformation setObject:keychainItem forKey:@"keychainItem"];
			}
			[[NSUserDefaults standardUserDefaults] setObject:username forKey:@"username"];
			[(PWindow*)loginWindow endSheet:self];
		}
		[self updateStations];
		
		NSString* startupStation = [[NSUserDefaults standardUserDefaults] objectForKey:@"startupStation"];
		if(startupStation == nil)
		{
			startupStation = [[stations objectAtIndex:0] objectForKey:@"id"];
			[[NSUserDefaults standardUserDefaults] setObject:startupStation forKey:@"startupStation"];
		}
		BOOL set = NO;
		int i;
		for(i = 0; i < [stations count]; i++)
		{
			if([[[stations objectAtIndex:i] objectForKey:@"id"] isEqual:startupStation])
			{
				[[stations objectAtIndex:i] setObject:[NSNumber numberWithInt:1] forKey:@"startup"];
				set = YES;
			}
		}
		if(!set)
		{
			startupStation = [[stations objectAtIndex:0] objectForKey:@"id"];
			[[stations objectAtIndex:0] setObject:[NSNumber numberWithInt:1] forKey:@"startup"];
		}
		
		[_pianobar selectStationById:startupStation];
		[_pianobar startStreaming];
		[volumeSlider setIntValue:0];
		[self updatePlaying];
	}
}

-(IBAction)changeStartupStation:(id)sender
{
	[self willChangeValueForKey:@"stations"];
	int i;
	for(i = 0; i < [stations count]; i++)
	{
		if([[[stations objectAtIndex:i] objectForKey:@"startup"] boolValue])
		{
			[[stations objectAtIndex:i] setObject:[NSNumber numberWithInt:0] forKey:@"startup"];
		}
	}
	[[[startupStationsController selectedObjects] objectAtIndex:0] setObject:[NSNumber numberWithInt:1] forKey:@"startup"];
	
	NSString* _id = [[[startupStationsController selectedObjects] objectAtIndex:0] objectForKey:@"id"];
	
	if(_id != nil)
		[[NSUserDefaults standardUserDefaults] setObject:_id forKey:@"startupStation"];
}

-(IBAction)hideStations:(id)sender
{
	int newHeight = 110;
	NSRect frame = [window frame];
	NSSize maxSize = NSMakeSize(FLT_MAX, FLT_MAX);
	NSSize minSize = NSMakeSize(300, 200);
	
	BOOL hide = (sender != self) ? ![(NSButton*)sender state] : YES;
	
	//NSLog(@"start: %f", frame.origin.y + frame.size.height);
	
	if(hide)
	{
		[stationsView setHidden:YES];
		[stationsView setAutoresizesSubviews:NO];
		
		int temp = 0;
		
		[[NSUserDefaults standardUserDefaults] setObject:[NSNumber numberWithInt:frame.size.height] forKey:@"window_height"];
		
		//FIXME: Hardcoded window size
		if(sender == self)
			temp = 112;
		
		[window setFrame:NSMakeRect(frame.origin.x, frame.origin.y + (frame.size.height - newHeight) - temp, frame.size.width, newHeight) display:YES animate:YES];
		[window setMaxSize:NSMakeSize(maxSize.width, newHeight)];
		[window setMinSize:NSMakeSize(minSize.width, newHeight)];
	}
	else
	{
		NSTableView* stationsTable = (NSTableView*)[[stationsView subviews] objectAtIndex:0];
		
		[stationsTable setFrame:NSMakeRect([stationsTable frame].origin.x, [stationsTable frame].origin.y, frame.size.width + 4, [stationsTable frame].size.height)];
		
		int oldheight;
		if([[NSUserDefaults standardUserDefaults] objectForKey:@"window_height"] != nil)
			oldheight = [[[NSUserDefaults standardUserDefaults] objectForKey:@"window_height"] intValue];
		else
			oldheight = minSize.height;
		
		[window setFrame:NSMakeRect(frame.origin.x, frame.origin.y + (frame.size.height - oldheight), frame.size.width, oldheight) display:YES animate:YES];
		[window setMaxSize:maxSize];
		[window setMinSize:minSize];
		[stationsView setAutoresizesSubviews:YES];
		[stationsView setHidden:NO];
	}
}

-(IBAction)deleteStation:(id)sender
{
	NSAlert* alert = [[[NSAlert alloc] init] autorelease];
	[alert setMessageText:[NSString stringWithFormat:@"Are you sure you want to delete the station \"%@\" ?", 
						   [[[stationsController selectedObjects] objectAtIndex:0] objectForKey:@"name"]]];
	[alert setInformativeText:@"This action cannot be undone."];
	[alert addButtonWithTitle:@"Cancel"];
	[alert addButtonWithTitle:@"OK"];
	if([alert runModal] == 1001)
	{
		NSString* _id = [[[stationsController selectedObjects] objectAtIndex:0] objectForKey:@"id"];
		[_pianobar deleteStationWithId:_id];
		[self updateStations];
	}
}

-(IBAction)addStation:(id)sender
{
	NSString* _id = [[[searchResultsController selectedObjects] objectAtIndex:0] objectForKey:@"id"];
	[_pianobar createStationById:_id];
	[self willChangeValueForKey:@"searchResults"];
	[searchResults removeAllObjects];
	[self didChangeValueForKey:@"searchResults"];
	[addStationWindow performClose:self];
	[self updateStations];
}

-(IBAction)pausePlay:(id)sender
{
	[_pianobar playpause];
	
	[self updatePlaying];
}

-(IBAction)next:(id)sender
{
	[_pianobar nextSong];
	[self clearInformation];
}


-(IBAction)loveSong:(id)sender
{
	if([_pianobar loveSong])
	{
		[self willChangeValueForKey:@"songInformation"];
		[songInformation setObject:[NSNumber numberWithInt:1] forKey:@"loveit"];
		[self didChangeValueForKey:@"songInformation"];
	}
}

-(IBAction)banSong:(id)sender
{
	[_pianobar banSong];
	[self clearInformation];
}

-(IBAction)explainSong:(id)sender
{
	NSAlert* alert = [[[NSAlert alloc] init] autorelease];
	[alert addButtonWithTitle:@"OK"];
	[alert setMessageText:[NSString stringWithFormat:@"Explanation For \"%@\"", [songInformation objectForKey:@"title"]]];
	[alert setInformativeText:[_pianobar getSongExplanation]];
	[alert runModal];
}

-(IBAction)search:(id)sender
{
	NSSearchField* field = (NSSearchField*)sender;
	[self willChangeValueForKey:@"searchResults"];
	[searchResults removeAllObjects];
	BOOL artist = [[(PSearchField*)sender searchField] isEqual:@"Artist"];
	if(![[field stringValue] isEqualToString:@""])
		[searchResults addObjectsFromArray:[_pianobar searchForMusicQuery:[field stringValue] isArtist:artist]];
	[self didChangeValueForKey:@"searchResults"];
}

-(IBAction)cancelLogin:(id)sender
{
	[(PWindow*)loginWindow endSheet:self];
	[[NSApplication sharedApplication] terminate:self];
}

-(IBAction)logout:(id)sender
{
	[self clearInformation];
	[loginInformation setObject:@"" forKey:@"username"];
	[loginInformation setObject:@"" forKey:@"password"];
	[[NSUserDefaults standardUserDefaults] removeObjectForKey:@"username"];
	[_pianobar stopStreaming];
	[(PWindow*)loginWindow showAsSheet:self];
}

@end
